// From the software distribution accompanying the textbook
// "A Practical Introduction to Data Structures and Algorithm Analysis,
// Third Edition (C++)" by Clifford A. Shaffer.
// Source code Copyright (C) 2007-2011 by Clifford A. Shaffer.

#include "dictionary.h"
#include "KVpair.h"

// Dictionary implemented with a hash table
// I re-implemented a bit of this in front of the class, so:
// TODO: 5 points on an assignment to anyone who can find an actual bug!

template <typename Key, typename E>
class hashdict: public Dictionary<Key, E>
{
    private:
        KVpair<Key, E> *HT;   // The hash table

        int M;          // Size of HT
        int currCnt;    // The current number of elements in HT
        int c = 3;      // constant for p(k, i) = ci

        Key EMPTYKEY;   // User-supplied key value for an empty slot
        Key TOMB;       // user will provide this lazy delet flag value


        int pLin(Key K, int i) const // Probe sequence
        {
            return i;
        }


        int pC(Key K, int i) const // Probe sequence
        {
            return c * i;
        }


        int pQuad(Key K, int i) const // Probe sequence
        {
            return i ^ 2;
        }


        // Note: these are overloaded, so the compiler will pick correctly

        // Simple hash function
        int h(int x) const
        {
            return x % M;
        }


        // Hash function for character array keys
        int h(char *x) const
        {
            int i, sum;

            for(sum = 0, i = 0; x[i] != '\0'; i++)
            {
                sum += (int)x[i];
            }

            return sum % M;
        }


        void hashInsert(const Key &, const E &);


        E hashSearch(const Key &) const;


    public:
        hashdict(int sz, Key k, Key z)
        {
            M = sz;
            EMPTYKEY = k;               // "k" is empty slot
            TOMB = z;

            currCnt = 0;

            HT = new KVpair<Key, E>[sz]; // Make HT of size sz

            for(int i = 0; i < M; i++)
            {
                HT[i].setKey(EMPTYKEY); // Initialize HT
            }
        }


        ~hashdict()
        {
            delete[] HT;
        }


        // Number stored in table
        int size()
        {
            return currCnt;
        }


        // Clear the dictionary
        void clear()
        {
            for(int i = 0; i < M; i++)
            {
                HT[i].setKey(EMPTYKEY);
            }

            currCnt = 0;
        }


        // Insert element "it" with Key "k" into the dictionary.
        void insert(const Key &k, const E &it)
        {
            Assert(currCnt < M, "Hash table is full");

            hashInsert(k, it);
            currCnt++;
        }


        // Find some record with key value "K"
        E find(const Key &k) const
        {
            return hashSearch(k);
        }


        // Remove was implemented while in front of class,
        // so it could be buggy
        E remove(const Key &k)
        {
            int home;                   // Home position for e
            int pos = home = h(k);      // Init probe sequence
            int i = 1;

            // while or for both work here
            // Leaves pos at the key or empty
            while(k != HT[pos].key() && HT[pos].key() != EMPTYKEY)
            {
                pos = (home + pLin(k, i)) % M; // probe
                i++;
            }

            // If pos indexes the key (it is not empty)
            if(HT[pos].key() != EMPTYKEY)
            {
                HT[pos].setKey(TOMB);
                currCnt--;

                return HT[pos].value();
            }
            else
            {
                return nullptr;
            }
            // Ask class:
            // Why is this ok after tomb?
            // What happens to these values?
            // Does it matter?
        }


        E removeAny()
        {   // Remove the first element occuring in the array
            // should not be used in any case but removing all
            // Ask class: Why is this needed?
            Assert(currCnt != 0, "Hash table is empty");

            int i;

            for(i = 0; i < M; i++)
            {
                if(HT[i].key() != EMPTYKEY && HT[i].key() != TOMB)
                {
                    HT[i].setKey(TOMB);
                    currCnt--;
                    break;
                }
            }

            return HT[i].value();
        }
};


// Insert e into hash table HT
template <typename Key, typename E>
void hashdict<Key, E>::hashInsert(const Key &k, const E &e)
{
    int home;                     // Home position for e
    int pos = home = h(k);        // Init probe sequence

    for(int i = 1; 
        EMPTYKEY != HT[pos].key() && TOMB !=HT[pos].key();
        i++)
    {
        Assert(k != HT[pos].key(), "Duplicates not allowed");

        // probe sequence
        pos = (home + pLin(k, i)) % M;
    }

    KVpair<Key,E> temp(k, e);
    HT[pos] = temp;
}


// Search for the record with Key K
template <typename Key, typename E>
E hashdict<Key, E>::hashSearch(const Key &k) const
{
    int home;                   // Home position for k
    int pos = home = h(k);      // Initial position is home slot

    // Leaves the pos index at the key location or empty location
    for(int i = 1;
        (k != HT[pos].key()) && (EMPTYKEY != HT[pos].key());
        i++)
    {
        pos = (home + pLin(k, i)) % M; // Next on probe sequence
    }

    if(k == HT[pos].key())      // Found it
    {
        // Returns the value stored under the searched key
        return HT[pos].value();
    }
    else
    {
        return nullptr;            // k not in hash table
        // this is why we are using a pointer type as the value in main
    }
}

